function regexpQuote(str) {
  return str.replace(/\W/g, $0 => '\\' + $0);
}

/*
 * 情報源を表す抽象オブジェクト (抽象メソッド: enumerate)
 * 情報源と文字列から CompletionResult を生成
 * CompletionResult は候補一覧や確定部分を持つ
 */
function CompletionSource(source, converter) {
  this.source = source;
  this.converter = converter || null;
}
//CompletionSource.kfalse = function (f) function (x) f(x), false;
CompletionSource.kfalse = f => x => (f(x), false);
CompletionSource.prototype.forEach = function (f) {
  this.enumerate(CompletionSource.kfalse(f));
};
CompletionSource.prototype.find = function (f) {
  let result = null;
  this.enumerate(function (i) {
    if (f(i)) {
      result = i;
      return true;
    }
    else {
      return false;
    }
  });
  return result;
};
CompletionSource.prototype.filter = function (f) {
  let ary = [];
  this.forEach(function (i) {
    if (f(i)) {
      ary.push(i);
    }
  });
  return ary;
};
CompletionSource.prototype.exists = function (str) {
  return Boolean(this.find(str));
};
CompletionSource.prototype.makeRegexp = function (str) {
  return this.converter ? this.converter(str) : new RegExp("^" + regexpQuote(str));
};
CompletionSource.prototype.complete = function (str) {
  return new CompletionResult(str, this);
};

function CompletionSourceArray(source, converter) {
  CompletionSource.call(this, source, converter);
}
// CompletionSourceArray.prototype.__proto__ = CompletionSource.prototype;
inherits(CompletionSourceArray, CompletionSource);
CompletionSourceArray.prototype.enumerate = function (f) {
  for (let i = 0; i < this.source.length; i++) {
    if (f(this.source[i])) break;
  }
};

function CompletionSourceKeys(source, converter) {
  CompletionSource.call(this, source, converter);
}
// CompletionSourceKeys.prototype.__proto__ = CompletionSource.prototype;
inherits(CompletionSourceKeys, CompletionSource);
CompletionSourceKeys.prototype.enumerate = function (f) {
  for (let i in this.source) {
    if (f(i)) break;
  }
};
CompletionSourceKeys.prototype.exists = function (str) {
  return Boolean(this.source[str]);
};

function CompletionSourceValues(source, converter) {
  CompletionSource.call(this, source, converter);
}
// CompletionSourceValues.prototype.__proto__ = CompletionSource.prototype;
inherits(CompletionSourceValues, CompletionSource);
CompletionSourceValues.prototype.enumerate = function (f) {
  for (let i of this.source) {
    if (f(i)) break;
  }
};

function commonPrefix(ary) {
  return ary.reduce(function (s, t) {
    let len = s.length < t.length ? s.length : t.length;
    let i;
    for (i = 0; i < len && s[i] == t[i]; i++);
    return s.substring(0, i);
  });
}
/* 候補の一覧，共通prefix */
/*
 * 候補一覧の生成 -> common prefix -> 補完
 * source     :: CompletionSource
 * candidates :: Array
 *     全候補
 * finished   :: Boolean
 *     結果が一意に定まれば true
 * value      :: String || null
 *     補完結果文字列 or prefix or マッチなしを表すnull
 */

function CompletionResult(str, source) {
  this.source = source;
  this.origin = str;

  let reg = source.makeRegexp(str);
  this.candidates = source.filter(s => reg.test(s));
  let len = this.candidates.length;
  if (len == 0) {
    this.finished = false;
    this.value = null;
  }
  else if (len == 1) {
    this.finished = true;
    this.value = this.candidates[0];
  }
  else {
    this.finished = false;
    let prefix = commonPrefix(this.candidates);
    this.value = reg.test(prefix) ? prefix : null;
  }
}
CompletionResult.prototype.exists = function () {
  return this.finished || this.source.exists(this.value);
};

// function complete(target, source) {
//   let result = source.complete(target.value);
//   let prefix = result.value;
//   let converter = source.converter;
//   if (prefix && (!converter || converter(target.value).test(prefix))) {
//     target.value = prefix;
//   }
//   return result;
// }

function CompletionHandler(target, src, show, clear, isCompletionDwimEvent, isCompletionEvent) {
  this.target = target;
  if (src) { this.source = src; }
  if (show) { this.show = show; }
  if (clear) { this.clear = clear; }
  if (isCompletionDwimEvent) {
    this.isCompletionDwimEvent = isCompletionDwimEvent;
  }
  if (isCompletionEvent) {
    this.isCompletionEvent = isCompletionEvent;
  }
  this._on = false;

  let self = this;
  let previousValue;
  this.keydown = function (ev) {
    previousValue = self.target.value;
  };
  this.keyup = function (ev) {
    let value = self.target.value;
    if (value && previousValue != value) {
      self.show(self.source.complete(value), self.target, self.source);
    }
  };
  this.keypress = function (ev) {
    if (self.isCompletionDwimEvent(ev)) {
      let value = self.target.value;
      if (!(value === "" || self.source.exists(value))) {
        let result = self.source.complete(value);
        if (result.value) {
          self.target.value = result.value;
        }
        if (!result.finished) {
          ev.preventDefault();
          return;
        }
      }
      self.clear();
    }
    else if (self.isCompletionEvent(ev)) {
      let value = self.target.value;
      let result = self.source.complete(value);
      if (result.value) {
        self.target.value = result.value;
      }
      self.show(result, self.target, self.source);
      ev.preventDefault();
    }
  };
}
CompletionHandler.prototype.show = function () {};
CompletionHandler.prototype.clear = function () {};
CompletionHandler.prototype.isCompletionDwimEvent = function (ev) {
  return ev.key === "Tab";
};
CompletionHandler.prototype.isCompletionEvent = function (ev) {
  return false;
};
CompletionHandler.prototype.on = function () {
  if (!this._on) {
    this.target.addEventListener("keydown", this.keydown, false);
    this.target.addEventListener("keyup", this.keyup, false);
    this.target.addEventListener("keypress", this.keypress, false);
    this._on = true;
  }
};
CompletionHandler.prototype.off = function () {
  this.target.deleteEventListener("keydown", this.keydown, false);
  this.target.deleteEventListener("keyup", this.keyup, false);
  this.target.deleteEventListener("keypress", this.keypress, false);
  this._on = false;
};

CompletionHandler.on = function (target, src, show) {
  (new CompletionHandler(target, src, show)).on();
};
